#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from hwt.hdlObjects.types.typeCast import toHVal
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.synthesizer.rtlLevel.rtlSignal import RtlSignal
from hwt.synthesizer.rtlLevel.signalUtils.cmp import areSameSignals


class Param(RtlSignal):
    """
    Class used in same way as generics in VHDL, it is wrapper around the value
    """
    
    def __init__(self, initval):
        initval = toHVal(initval)
        super(Param, self).__init__(None, None, initval._dtype, defaultVal=initval)
        self._val = initval
        self.replacedWith = None
        self._parent = None
        # unit: (ctx, name)
        self._scopes = {}
    
    def _registerScope(self, name, unit):     
        self._scopes[unit] = (unit._cntx, name)
    
    def getName(self, where):
        return self._scopes[where][1]
        
    def setHdlName(self, name):
        self.hasGenericName = False
        self.name = name
        
    def get(self):
        if self.replacedWith is not None:
            raise Exception("Trying to read param '%s' which is already replaced by '%s'" % 
                            (str(self), str(self.replacedWith)))
        return self._val
  
    def replace(self, replaceWith):
        """
        self will always have value of parent
        """
        if areSameSignals(self, replaceWith):
            return
        
        if self.replacedWith is not None:
            raise Exception("replacing '%s' with '%s' and it was already replaced by '%s'" % 
                            (str(self), str(replaceWith), str(self.replacedWith)))
            
        for dr in self.drivers:
            dr.ops = [replaceWith if x is self else x for x in dr.ops]
                
        for ep in self.endpoints:
            ep.ops = [replaceWith if x is self else x for x in ep.ops]
            
        
        self.replacedWith = replaceWith
    

    def set(self, val):
        """
        set value of this param
        """
        assert self.replacedWith is None
        val = toHVal(val)
        self.defaultVal = val
        self._val = val.staticEval()
    
    def __repr__(self):
        val = "InvalidVal"
        try:
            val = self.get()
        except Exception:
            pass 
        
        try:
            name = self.name
        except AttributeError:
            name = ""
        
        return "<%s, name=%s, val=%s>" % (self.__class__.__name__, name, str(val)) 
        

def evalParam(p):
    """
    Get value of parameter
    """
    while isinstance(p, Param):
        p = p.get()
    
    if isinstance(p, RtlSignalBase):
        return p.staticEval()
        # use rather param inheritance instead of param as param value
    return toHVal(p)

def getParam(p):
    """
    get param value or value
    getParam gets only item which is specified as param value,
    but evalParam also evaluates it if it is not value 
    """
    if isinstance(p, Param):
        p = p.get()
        # use rather param inheritance instead of param as param value
    return p



from jinja2.environment import Environment
from jinja2.loaders import PackageLoader
from keyword import kwlist

from hwt.hdlObjects.operator import Operator
from hwt.hdlObjects.operatorDefs import AllOps, sensitivityByOp
from hwt.hdlObjects.constants import SENSITIVITY
from hwt.hdlObjects.statements import IfContainer
from hwt.hdlObjects.types.enum import Enum
from hwt.hdlObjects.types.enumVal import EnumVal
from hwt.hdlObjects.value import Value
from hwt.serializer.exceptions import SerializerException
from hwt.serializer.nameScope import LangueKeyword, NameScope
from hwt.serializer.simModel.value import SimModelSerializer_value
from hwt.serializer.simModel.ops import SimModelSerializer_ops
from hwt.serializer.simModel.types import SimModelSerializer_types
from hwt.serializer.utils import maxStmId
from hwt.synthesizer.param import evalParam
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.hdlObjects.types.bits import Bits


env = Environment(loader=PackageLoader('hwt', 'serializer/simModel/templates'))
unitTmpl = env.get_template('modelCls.py')
processTmpl = env.get_template('process.py')
ifTmpl = env.get_template("if.py")

_indent = "    "
_indentCache = {}        
def getIndent(indentNum):
    try:
        return  _indentCache[indentNum]
    except KeyError:
        i = "".join([_indent for _ in range(indentNum)])   
        _indentCache[indentNum] = i
        return i

class SimModelSerializer(SimModelSerializer_value, SimModelSerializer_ops, SimModelSerializer_types):
    __keywords_dict = {kw: LangueKeyword() for kw in kwlist}
    __keywords_dict.update({'sim': LangueKeyword(),
                            'self': LangueKeyword()})
    fileExtension = '.py'
    formater = lambda s: s
    
    @classmethod
    def getBaseNameScope(cls):
        s = NameScope(True)
        s.setLevel(1)
        s[0].update(cls.__keywords_dict)
        return s
    
    @classmethod
    def serializationDecision(cls, obj, serializedClasses, serializedConfiguredUnits):
        # we need all instances for simulation
        return True
    
    @classmethod
    def asHdl(cls, obj):
        if isinstance(obj, RtlSignalBase):
            return cls.SignalItem(obj)
        elif isinstance(obj, Value):
            return cls.Value(obj)
        else:
            try:
                serFn = getattr(cls, obj.__class__.__name__)
            except AttributeError:
                raise NotImplementedError("Not implemented for %s" % (repr(obj)))
            return serFn(obj)
    
    @classmethod
    def stmAsHdl(cls, obj, indent=0, enclosure=None):
        try:
            serFn = getattr(cls, obj.__class__.__name__)
        except AttributeError:
            raise NotImplementedError("Not implemented for %s" % (repr(obj)))
        return serFn(obj, indent, enclosure)
    
    @classmethod
    def FunctionContainer(cls, fn):
        raise NotImplementedError()
        # return fn.name
    @classmethod
    def Entity(cls, ent, scope):
        ent.name = scope.checkedName(ent.name, ent, isGlobal=True)
        return ""
        
    @classmethod
    def Architecture(cls, arch, scope):
        variables = []
        procs = []
        extraTypes = set()
        extraTypes_serialized = []
        arch.variables.sort(key=lambda x: x.name)
        arch.processes.sort(key=lambda x: (x.name, maxStmId(x)))
        arch.componentInstances.sort(key=lambda x: x._name)
        
        for v in arch.variables:
            t = v._dtype
            # if type requires extra definition
            if isinstance(t, Enum) and t not in extraTypes:
                extraTypes.add(v._dtype)
                extraTypes_serialized.append(cls.HdlType(t, scope, declaration=True))

            v.name = scope.checkedName(v.name, v)
            variables.append(v)
            
        
        def serializeVar(v):
            dv = evalParam(v.defaultVal)
            if isinstance(dv, EnumVal):
                dv = "%s.%s" % (dv._dtype.name, dv.val)
            else:
                dv = cls.Value(dv)
            
            return v.name, cls.HdlType(v._dtype), dv
        
        for p in arch.processes:
            procs.append(cls.HWProcess(p, scope, 0))
        
        # architecture names can be same for different entities
        # arch.name = scope.checkedName(arch.name, arch, isGlobal=True)    
             
        return unitTmpl.render({
        "name"               : arch.getEntityName(),
        "ports"              : list(map(lambda p: (p.name, cls.HdlType(p._dtype)), arch.entity.ports)),
        "signals"            : list(map(serializeVar, variables)),
        "extraTypes"         : extraTypes_serialized,
        "processes"          : procs,
        "processObjects"     : arch.processes,
        "processesNames"     : map(lambda p: p.name, arch.processes),
        "componentInstances" : arch.componentInstances,
        "isOp"               : lambda x: isinstance(x, Operator),
        "sensitivityByOp"    : sensitivityByOp
        })
   
    @classmethod
    def Assignment(cls, a, indent=0, default=None):
        dst = a.dst
        indentStr = getIndent(indent)
        ev = a.isEventDependent
        
        if a.indexes is not None:
            return "%syield (self.%s, %s, (%s,), %s)" % (
                        indentStr, dst.name, cls.Value(a.src),
                        ", ".join(map(cls.asHdl, a.indexes)), ev)
        else:
            if not (dst._dtype == a.src._dtype):
                srcT = a.src._dtype
                dstT = dst._dtype
                if (isinstance(srcT, Bits) and isinstance(dstT, Bits) and
                    srcT.bit_length() == dstT.bit_length() == 1):
                    if srcT.forceVector != dstT.forceVector:
                        if srcT.forceVector:
                            return "%syield (self.%s, (%s)._getitem__val(simHInt(0)), %s)" % (
                                    indentStr, dst.name, cls.Value(a.src), ev)
                        else:
                            return "%syield (self.%s, %s, (simHInt(0),), %s)" % (
                                    indentStr, dst.name, cls.Value(a.src), ev)
                    
                raise SerializerException(("%s <= %s  is not valid assignment\n" +
                                          " because types are different (%r; %r) ") % 
                                          (cls.asHdl(dst), cls.Value(a.src),
                                          dst._dtype, a.src._dtype))
            else:
                return "%syield (self.%s, %s, %s)" % (
                        indentStr, dst.name, cls.Value(a.src), ev)
            
    @classmethod
    def comment(cls, comentStr):
        return "#" + comentStr.replace("\n", "\n#")     

    @classmethod
    def condAsHdl(cls, cond):
        cond = list(cond)
        return "%s" % (",".join(map(lambda x: cls.asHdl(x), cond)))
    
    @classmethod
    def IfContainer(cls, ifc, indent, enclosure=None):
        cond = cls.condAsHdl(ifc.cond)
        ifTrue = ifc.ifTrue
        ifFalse = ifc.ifFalse

        if ifc.elIfs:
            # if has elifs revind this to tree
            ifFalse = []
            topIf = IfContainer(ifc.cond, ifc.ifTrue, ifFalse)
            for c, stms in ifc.elIfs:
                _ifFalse = []
                lastIf = IfContainer(c, stms, _ifFalse)
                ifFalse.append(lastIf)
                ifFalse = _ifFalse
            
            lastIf.ifFalse = ifc.ifFalse
            
            return cls.IfContainer(topIf, indent, enclosure)
        else:
            if enclosure is None:
                _enclosure = getIndent(indent + 1) + "pass"
            else:
                _enclosure = cls.stmAsHdl(enclosure, indent + 1)
                
            return ifTmpl.render(
                indent=getIndent(indent),
                indentNum=indent,
                cond=cond,
                enclosure=_enclosure,
                ifTrue=tuple(map(lambda obj: cls.stmAsHdl(obj, indent + 1, enclosure),
                                 ifTrue)),
                ifFalse=tuple(map(lambda obj: cls.stmAsHdl(obj, indent + 1, enclosure),
                                   ifFalse)))  
    
    @classmethod
    def SwitchContainer(cls, sw, indent, enclosure=None):
        switchOn = sw.switchOn
        mkCond = lambda c: {Operator(AllOps.EQ,
                                    [switchOn, c])}
        ifFalse = []
        elIfs = []
        
        for key, statements in sw.cases:
            if key is not None:  # None is default
                elIfs.append((mkCond(key), statements))  
            else:
                ifFalse = statements
        
        topCond = mkCond(sw.cases[0][0])
        topIf = IfContainer(topCond,
                    sw.cases[0][1],
                    ifFalse,
                    elIfs)
        
        return cls.IfContainer(topIf, indent, enclosure)
    
    @classmethod
    def WaitStm(cls, w):
        if w.isTimeWait:
            return "wait for %d ns" % w.waitForWhat
        elif w.waitForWhat is None:
            return "wait"
        else:
            raise NotImplementedError()
    
    @classmethod
    def sensitivityListItem(cls, item):
        if isinstance(item, Operator):
            op = item.operator
            if op == AllOps.RISING_EDGE:
                sens = SENSITIVITY.RISING
            elif op == AllOps.FALLIGN_EDGE:
                sens = SENSITIVITY.FALLING
            else:
                assert op == AllOps.EVENT
                sens = SENSITIVITY.ANY
            
            return "(%s, %s)" % (str(sens), item.ops[0].name)
        else:
            return item.name
    
    @classmethod
    def HWProcess(cls, proc, scope, indentLvl):
        body = proc.statements
        proc.name = scope.checkedName(proc.name, proc)
        sensitivityList = sorted(map(cls.sensitivityListItem, proc.sensitivityList))
        if len(body) == 1:
            _body = cls.stmAsHdl(body[0], 2)
        elif len(body) == 2:
            # first statement is taken as default
            _body = cls.stmAsHdl(body[1], 2, body[0])
        else:
            raise NotImplementedError()
        
        return processTmpl.render({
              "name": proc.name,
              "sensitivityList": sensitivityList,
              "stmLines": [_body] })
           





from hwt.hdlObjects.types.boolean import Boolean
from hwt.hdlObjects.types.integer import Integer
from hwt.hdlObjects.types.bits import Bits
from hwt.hdlObjects.types.string import String
from hwt.hdlObjects.types.slice import Slice



BOOL = Boolean()
INT = Integer()
UINT = Integer(_min=0)
PINT = Integer(_min=1)
BIT = Bits(widthConstr=1)
VECTOR = Bits(forceVector=True)
STR = String()    
SLICE = Slice()







from hwt.synthesizer.interfaceLevel.interfaceUtils.utils import walkPhysInterfaces


def defaultUnitName(unit, sugestedName=None):
    if not sugestedName:
        defNameFn = None
        try:
            defNameFn = unit._getDefaultName
        except AttributeError:
            pass
        
        if defNameFn:
            return defNameFn()
            
        return unit.__class__.__name__
    else:
        return sugestedName

def walkSignalOnUnit(unit):
    for i in unit._interfaces:
        yield from walkPhysInterfaces(i)



from hwt.interfaces.agents.signal import SignalAgent
from hwt.interfaces.agents.vldSynced import VldSyncedAgent
from hwt.simulator.agentBase import SyncAgentBase


class RegCntrlAgent(SyncAgentBase):
    """
    Simulation/verification agent for RegCntrl interface
    """
    def __init__(self, intf, clk=None, rstn=None):
        self.intf = intf
        if clk is None:
            clk = self._getClk()
        
        self._din = SignalAgent(intf.din, clk=clk, rstn=rstn)
        self._dout = VldSyncedAgent(intf.dout, clk=clk, rstn=rstn, allowNoReset=True)

    def din_getter(self):
        return self._din.data

    def din_setter(self, newVal):
        self._din.data = newVal
    
    din = property(din_getter, din_setter)
        
    def dout_getter(self):
        return self._dout.data

    def dout_setter(self, newVal):
        self._dout.data = newVal
    
    dout = property(dout_getter, dout_setter)
        
    def getDrivers(self):
        return self._din.getMonitors() + self._dout.getDrivers()
    
    def getMonitors(self):
        return self._din.getDrivers() + self._dout.getMonitors()


from hwt.hdlObjects.constants import Time
from hwt.simulator.agentBase import AgentBase
from hwt.simulator.shortcuts import oscilate


class OscilatorAgent(AgentBase):
    def __init__(self, intf, period=10 * Time.ns):
        self.period = period
        self.driver = oscilate(intf, period=period)
        



from hwt.hdlObjects.value import Value
from hwt.hdlObjects.function import Function
from hwt.hdlObjects.types.defs import INT
from operator import floordiv, add, sub, inv, mod, mul, ne, and_, or_, \
    xor, gt, ge, lt, le, getitem
from hwt.hdlObjects.constants import SENSITIVITY

class OpDefinition():
    """
    OperatorDefinition
    @ivar id: name of operator
    @ivar _evalFn: function which evaluates operands
    """
    def __init__(self, evalFn):
        self.id = None  # assigned automatically in AllOps  
        self._evalFn = evalFn
        
    def __eq__(self, other):
        return type(self) == type(other) and self.id == other.id
    
    def  __hash__(self):
        return hash(self.id)
    
    def eval(self, operator, simulator=None):
        """Load all operands and process them by self._evalFn"""
        def getVal(v):
            while not isinstance(v, (Value, Function)):
                v = v._val
            
            return v

        ops = list(map(getVal, operator.ops))
        
        if isEventDependentOp(operator.operator):
            ops.append(simulator.now)
        elif operator.operator == AllOps.IntToBits:
            ops.append(operator.result._dtype)
        
        return self._evalFn(*ops)

    def __repr__(self):
        return "<OpDefinition %s>" % (self.id)

def isEventDependentOp(operator):
    return operator in (AllOps.RISING_EDGE, AllOps.EVENT, AllOps.FALLIGN_EDGE)


def eventFn(a, now):
    return a._hasEvent(now)

def onRisingEdgeFn(a, now):
    return a._onRisingEdge(now)

def onFallingEdge(a, now):
    return a._onFallingEdge(now)

def dotOpFn(a, name):
    return getattr(a, name)

# [TODO] downto / to are relict of vhdl and should be replaced with slice
def downtoFn(a, b):
    return a._downto(b)

def toFn(a, b):
    return a._to(b)

def concatFn(a, b):
    return a._concat(b)

def power(base, exp):
    return base._pow(exp) 

def eqFn(a, b):
    return a._eq(b)

def ternaryFn(a, b, c):
    return a._ternary(b, c)

def callFn(fn, *ops):
    return a.call(*ops)

def bitsToIntFn(a):
    return a._convert(INT)

def intToBitsFn(a, t):
    return  a._convert(t)

def bitsAsSignedFn(a):
    return a._signed()

def bitsAsUnsignedFn(a):
    return a._unsigned()

def bitsAsVec(a):
    return a._vec()

class AllOps():
    _idsInited = False
    """
    @attention: Remember that and operator "and" is & and "or" is |, "and" and "or" can not be used because
    they can not be overloaded
    @attention: These are operators of internal AST, the are not equal to verilog or vhdl operators
    """
    
    
    EVENT = OpDefinition(eventFn)
    RISING_EDGE = OpDefinition(onRisingEdgeFn)  # unnecessary
    FALLIGN_EDGE = OpDefinition(onFallingEdge)  # unnecessary
    
    DIV = OpDefinition(floordiv)
    ADD = OpDefinition(add)
    SUB = OpDefinition(sub)
    POW = OpDefinition(power)
    UN_MINUS = OpDefinition(inv)
    MOD = OpDefinition(mod)
    MUL = OpDefinition(mul)
    
    NOT = OpDefinition(inv)
    XOR = OpDefinition(xor)
    AND_LOG = OpDefinition(and_)
    OR_LOG = OpDefinition(or_)

    DOT = OpDefinition(dotOpFn)
    DOWNTO = OpDefinition(downtoFn)
    TO = OpDefinition(toFn)
    CONCAT = OpDefinition(concatFn)
    
    EQ = OpDefinition(eqFn)
    NEQ = OpDefinition(ne)
    GREATERTHAN = OpDefinition(gt)
    GE = OpDefinition(ge)
    LOWERTHAN = OpDefinition(lt)
    LE = OpDefinition(le)
    

    INDEX = OpDefinition(getitem)
    TERNARY = OpDefinition(ternaryFn)
    CALL = OpDefinition(callFn)
    
    BitsToInt = OpDefinition(bitsToIntFn)
    IntToBits = OpDefinition(intToBitsFn)
    
    BitsAsSigned = OpDefinition(bitsAsSignedFn)
    BitsAsUnsigned = OpDefinition(bitsAsUnsignedFn)
    BitsAsVec = OpDefinition(bitsAsVec)
    
    allOps = {}
        
    @classmethod
    def opByName(cls, name):
        return getattr(cls, name)
    
if not AllOps._idsInited:
    for a in dir(AllOps):
        o = getattr(AllOps, a)
        if isinstance(o, OpDefinition):
            o.id = a
            
    AllOps._idsInited = True

def sensitivityByOp(op):
    if op == AllOps.RISING_EDGE:
        return SENSITIVITY.RISING
    elif op == AllOps.FALLIGN_EDGE:
        return SENSITIVITY.FALLING
    else:
        assert op == AllOps.EVENT
        return SENSITIVITY.ANY



from hwt.synthesizer.param import evalParam
from hwt.hdlObjects.types.sliceVal import SliceVal
from hwt.hdlObjects.types.bits import Bits
from hwt.hdlObjects.types.enum import Enum
from hwt.hdlObjects.types.array import Array
from hwt.hdlObjects.types.hdlType import HdlType
from hwt.serializer.exceptions import SerializerException
from hwt.hdlObjects.types.integer import Integer

class SimModelSerializer_types():
    @classmethod
    def HdlType_bits(cls, typ, declaration=False):
        if typ.signed is None:
            if not (typ.forceVector or typ.bit_length() > 1):
                return 'BIT'
            
        c = typ.constrain
        if isinstance(c, (int, float)):
            pass
        else:        
            c = evalParam(c)
            if isinstance(c, SliceVal):
                c = c._size()
            else:
                c = c.val  
             
        return "simBitsT(%d, %r)" % (c, typ.signed)

    @classmethod
    def HdlType_enum(cls, typ, scope, declaration=False):
        if declaration:
            try:
                name = typ.name
            except AttributeError:
                name = "enumT_"
            typ.name = scope.checkedName(name, typ)
            
            return '%s = Enum( "%s", [%s])' % (typ.name, typ.name, ", ".join(map(lambda x: '"%s"' %x, typ._allValues)))
        else:
            return typ.name
        
    @classmethod
    def HdlType_int(cls, typ, scope, declaration=False):
        ma = typ.max
        mi = typ.min
        noMax = ma is None
        noMin = mi is None
        if noMin: 
            if noMax:
                return "SIM_INT"
            else:
                raise SerializerException("If max is specified min has to be specified as well")
        else:
            if noMax:
                # [TODO] convert these to sim as well
                if mi == 0:
                    return "UINT"
                elif mi == 1:
                    return "PINT"
                else:
                    raise SerializerException("If max is specified min has to be specified as well")
            else:
                raise NotImplementedError()

    @classmethod
    def HdlType_array(cls, typ, scope, declaration=False):
        assert not declaration
        return "Array(%s, %d)" % (cls.HdlType(typ.elmType), evalParam(typ.size).val)

    @classmethod
    def HdlType(cls, typ, scope=None, declaration=False):
        if isinstance(typ, Bits):
            return cls.HdlType_bits(typ, declaration=declaration)
        elif isinstance(typ, Enum):
            return cls.HdlType_enum(typ, scope, declaration=declaration)
        elif isinstance(typ, Array):
            return cls.HdlType_array(typ, scope, declaration=declaration)
        elif isinstance(typ, Integer):
            return cls.HdlType_int(typ, scope, declaration=declaration)
        else:
            if declaration:
                raise NotImplementedError("type declaration is not implemented for type %s" % 
                                      (typ.name))
            else:
                assert isinstance(typ, HdlType)
                return typ.name.upper()




class HdlRef():
    """
    hdl reference container
    """
    def __init__(self, names, caseSensitive, allChilds=False):
        self.names = HdlRef._caseSensitivityForNames(caseSensitive, names)
        self.all = allChilds

    @staticmethod
    def _caseSensitivityForNames(caseSensitive, names):
        if caseSensitive:
            return names
        else:
            return tuple([n.lower() for n in names])

    def __str__(self):
        SEPARATOR = '.'
        s = SEPARATOR.join(self.names)
        if self.all:
            s += SEPARATOR + "all"
        return s

    def __repr__(self):
        return "<HdlRef  " + str(self) + ">"



import itertools
from hwt.hdlObjects.types.integer import Integer
from hwt.hdlObjects.operator import Operator
from hwt.hdlObjects.operatorDefs import AllOps

class FunctionContainer(list):
    """
    Hdl container for functions with same name to support function overloading
    """
    def __init__(self, name, parent):
        super(FunctionContainer, self).__init__()
        self.name = name
        self.parent = parent
    
    def __hash__(self):
        return hash((id(self), self.name))
        
    def append(self, fn, suppressRedefinition=False):
        """
        Append function definition
        """
        if self:
            assert self[0].name == fn.name  # assert every appended function has same name
        if not suppressRedefinition:
            for _fn in self:  # check if same definition exists
                same = True
                for _p, p in itertools.zip_longest(_fn.params, fn.params):
                    if _p._dtype != p._dtype:
                        same = False
                        break
                assert not same

        return list.append(self, fn)

    def lookup(self, *args):
        """
        lookup function definition by args
        this is called on body (f.e. function implementation vhdl package body)
        """
        # check if same definition exists
        for fn in self:
            same = True
            for _p, p in itertools.zip_longest(args, fn.params):
                if _p._dtype != p._dtype and \
                 not (isinstance(_p._dtype, Integer), isinstance(p._dtype, Integer)):
                    same = False
                    break
            if same:
                return fn
        raise KeyError("Function was not found")
    
    def call(self, *args):
        # this is called on header
        body = self.parent.body[self.name]
        
        fn = body.lookup(*args)
        return Operator.withRes(AllOps.CALL, [fn, *args], fn.returnT)
        
    def staticEval(self):
        # function id does not have to be evalueated
        pass
    



from hwt.hdlObjects.value import Value
from hwt.hdlObjects.types.integer import Integer
from hwt.hdlObjects.types.bits import Bits
from hwt.hdlObjects.types.hdlType import HdlType
from hwt.hdlObjects.operator import Operator
from hwt.hdlObjects.operatorDefs import AllOps
from hwt.hdlObjects.types.defs import BOOL
from hwt.hdlObjects.types.booleanVal import BooleanVal


def convertInteger__val(self, sigOrVal, toType):
    if isinstance(toType, Integer):
        v = sigOrVal.clone()
        if toType.min is not None:
            assert v.val >= toType.min
        if toType.max is not None:
            assert v.val <= toType.max
                
        v._dtype = toType
        return v
    elif toType == BOOL:
        v = sigOrVal.val
        assert v == 0 or v == 1
        return BooleanVal(v, BOOL, sigOrVal.vldMask, sigOrVal.updateTime)
            
    elif isinstance(toType, Bits):
        _v = sigOrVal.val 
        w = toType.bit_length()
        assert _v.bit_length() <= w, (_v, w)
        v = toType.fromPy(_v)
        
        v.updateTime = sigOrVal.updateTime
         
        v._dtype = toType
        if not sigOrVal.vldMask:
            v.vldMask = 0
        return v
            
    return HdlType.defaultConvert(self, sigOrVal, toType)



def convertInteger(self, sigOrVal, toType):
    isVal = isinstance(sigOrVal, Value)
    if isinstance(toType, Integer):
        if isVal:
            v = sigOrVal.clone()
            if toType.min is not None:
                assert v.val >= toType.min
            if toType.max is not None:
                assert v.val <= toType.max
                
            v._dtype = toType
            return v
        else:
            return sigOrVal  # [TODO] use convertor op
    elif toType == BOOL:
        if isVal:
            v = sigOrVal.val
            assert v == 0 or v == 1
            return BooleanVal(v, BOOL, sigOrVal.vldMask, sigOrVal.updateTime)
            
    elif isinstance(toType, Bits):
        if isVal:
            _v = sigOrVal.val 
            w = toType.bit_length()
            assert _v.bit_length() <= w, "%d can not fit into %d bits" % (_v, w)
            v = toType.fromPy(_v)
            
            v.updateTime = sigOrVal.updateTime
             
            v._dtype = toType
            if not sigOrVal.vldMask:
                v.vldMask = 0
            return v
        else:
            return Operator.withRes(AllOps.IntToBits, [sigOrVal], toType)

            
    return HdlType.defaultConvert(self, sigOrVal, toType)



from hwt.hdlObjects.variables import SignalItem
from hwt.hdlObjects.constants import DIRECTION

class PortItem(SignalItem):
    """basic hdl entity port item"""
    def __init__(self, name, direction, dtype, unit):
        self.name = name
        self.unit = unit
        self.direction = direction
        self._dtype = dtype
        self.src = None
        self.dst = None
    
    
    def simEval(self, simulator):
        """
        dst signal is updated when src signal is updated no specific reevaluating is needen
        """
        pass
    
    
    def connectSig(self, signal):
        """
        Connect to port item on subunit
        """
        if self.direction == DIRECTION.IN:
            if self.src is not None:
                raise Exception("Port %s is already associated with %s" % (self.name, str(self.src)))
            self.src = signal
            
            signal.endpoints.append(self)
        elif self.direction == DIRECTION.OUT:
            if self.dst is not None:
                raise Exception("Port %s is already associated with %s" % (self.name, str(self.dst)))
            self.dst = signal
            
            signal.drivers.append(self)
        else:
            raise NotImplementedError()
    
        signal.hidden = False
        signal.ctx.subUnits.add(self.unit)
    
    def reigsterInternSig(self, signal):
        """
        Connect internal signal to port item,
        this connection is used by simulator and only output port items will be connected
        """
        if self.direction == DIRECTION.OUT:
            if self.src is not None:
                raise Exception("Port %s is already associated with %s" % (self.name, str(self.src)))
            self.src = signal
        elif self.direction == DIRECTION.IN:
            if self.dst is not None:
                raise Exception("Port %s is already associated with %s" % (self.name, str(self.dst)))
            self.dst = signal
        else:
            raise NotImplementedError()

    def connectInternSig(self):
        """
        connet signal from internal side of of this component to this port
        """
        if self.direction == DIRECTION.OUT:
            self.src.endpoints.append(self)
        elif self.direction == DIRECTION.IN:
            self.dst.drivers.append(self)
        elif self.direction == DIRECTION.INOUT:
            self.dst.drivers.append(self)
        else:
            raise NotImplementedError()
        
    def __repr__(self):
        from hwt.serializer.vhdl.serializer import VhdlSerializer, onlyPrintDefaultValues
        return VhdlSerializer.PortItem(self, onlyPrintDefaultValues) 




from hwt.hdlObjects.operator import Operator
from hwt.hdlObjects.operatorDefs import AllOps
from hwt.hdlObjects.types.defs import BOOL
from hwt.hdlObjects.value import Value


class EventCapableVal(Value):
    def _hasEvent__val(self, now):
        BoolVal = BOOL.getValueCls()
        return BoolVal(self.updateTime == now,
                            BOOL,
                            self.vldMask,
                            now)
        
    def _hasEvent(self, now):
        if isinstance(self, Value):
            return self._hasEvent__val(now)
        else:
            return Operator.withRes(AllOps.EVENT, [self], BOOL)
    
    def _onFallingEdge__val(self, now):
        v = self._hasEvent__val(now)
        v.val = v.val and not self.val
        return v
    def _onFallingEdge(self, now):
        if isinstance(self, Value):
            return self._onFallingEdge__val(now)
        else:
            return Operator.withRes(AllOps.FALLIGN_EDGE, [self], BOOL)
    
    def _onRisingEdge__val(self, now):
        v = self._hasEvent__val(now)
        v.val = v.val and self.val
        return v   
    def _onRisingEdge(self, now):
        if isinstance(self, Value):
            return self._onRisingEdge__val(now)
        else:
            return Operator.withRes(AllOps.RISING_EDGE, [self], BOOL)